import torch
import torch.nn as nn
import torch.nn.functional as F

import numpy as np
import matplotlib.pyplot as plt
# import os

torch.cuda.empty_cache()
π = np.pi
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

#############################################################
### BASIC STUFF #############################################
#############################################################

def nn_size(nn: nn.Module): return sum([p.numel() for p in nn.parameters()]) # get param count

def nn_freeze(model: nn.Module):
	model.eval() # TODO : prevent going back to train mode
	for param in model.parameters(): param.requires_grad = False
	for l in model.children(): nn_freeze(l)
	return model

def nn_unfreeze(model: nn.Module):
	for param in model.parameters(): param.requires_grad = True
	for l in model.children(): nn_freeze(l)
	return model

def optim_lr(optimizer: torch.optim.Optimizer, lr: float):
	for g in optimizer.param_groups: g['lr'] = lr

#############################################################
### LOSSES ##################################################
#############################################################

# you want this to tend to 1. If you try to increase this, at least the direction will be right, like cosine similarity
class CosLoss(nn.Module):
	def __init__(self):
		super(CosLoss,self).__init__()
		self.cossim = nn.CosineSimilarity()
	def forward(self,x,y):
		return 1 - self.cossim(x,y).mean()

# you want this to tend to 1. If you try to increase this, at least the direction will be right, like cosine similarity
class DotSim(nn.Module):
	def __init__(self):
		super(DotSim,self).__init__()
	def forward(self,x,y):
		# y is interpreted as a target, we normalize against y only :
		# y -> y/|y|
		# x -> x/|y|
		# thus we return ⟨x|y⟩/|y|²
		yns = (y*y).sum(dim=1).unsqueeze(1) # y_norm_squared
		return (x*y/yns).sum(dim=1)

# you want to minimize this
class DotLoss(nn.Module):
	def __init__(self):
		super(DotLoss,self).__init__()
		self.dot=DotSim()
		self.mse=nn.MSELoss()
	def forward(self,x,y):
		return self.mse(self.dot(x,y),self.dot(y,y))

#############################################################
### MASK FUNCTIONS ##########################################
#############################################################

def checkerboard(h,w):
	return torch.stack(torch.meshgrid(torch.arange(h), torch.arange(w))).sum(axis=0) % 2

def mask_flatten(θ,mask):
	f = nn.Flatten()
	θθ = f(θ)[f(mask)]
	return θθ

# TODO : performance wise, should we gen a tensor ? (bad idea in term of memory ?)
#def mask_unflatten(θθ,mask,default_value = 0.):
#  f = nn.Flatten()
#  m = f(mask)
#  m = m[0] if len(m.shape)>1 else m
#  θ = torch.zeros(m.shape)
#  c = 0
#  for i,j in enumerate(m):
#    j = j.item()
#    θ[i] = θθ[c] if j else 0
#    c+=j
#  θ = θ.view(mask.shape)
#  return θ

def mask_unflatten(θθ,mask,default_value = 0.):
	N = θθ.shape[0]
	f = nn.Flatten()
	m = f(mask)
	m = m[0] if len(m.shape)>1 else m
	θ = torch.zeros((N,)+m.shape).to(θθ.device)
	c = 0
	for i,j in enumerate(m):
		j = j.item()
		θ[:,i] = θθ[:,c] if j else 0
		c+=j
	θ = θ.view((N,)+mask.shape)
	return θ

""" usage/debug :
resx=resy=28
mask = (1 - F.pad(SOLO_conv(torch.ones(1,resy,resx),stride=2), [1,1,1,1], "constant", 0)).bool().to(device)
N=10
x = mnist_set_train_x[:N].to(device) # torch.randn(N,1,resy,resx).to(device)
θ = torch.randn(1,2*resy+1,2*resx+1).to(device)
y = torch.randn(N,10) # φ(x,θ)

θθ = mask_flatten(θ,mask).repeat(N,1)
θ*mask = mask_unflatten(θθ,mask) 
"""

""" usage :
  mask = torch.randn(1,5,5)<0
  θ = torch.randn(1,5,5)*mask

  θθ = mask_flatten(θ,mask)
  θ = mask_unflatten(θθ,mask)
"""

#############################################################
### TENSOR MAKER ############################################
#############################################################

def kernel_from_fun2d(fun2d,Nx,Ny):
	A = np.fromfunction(lambda j, i: fun2d(2*i/(Nx-1)-1,2*j/(Ny-1)-1), (Ny,Nx))
	return torch.tensor(A,dtype=torch.float32)[None,None,:,:]

def kernel_gauss(Nx,Ny,σ=1):
	return kernel_from_fun2d(lambda x,y: np.exp(-(x*x+y*y)/(σ**2)),Nx,Ny)

#############################################################
### COMPLEX NN ##############################################
#############################################################

class Linear(nn.Module):
	def __init__(self,dim_i,dim_o,bias=True):
		super(Linear,self).__init__()
		self.linear = nn.Linear(dim_i,dim_o,bias,dtype=torch.cfloat)
	def forward(self,x):
		# l = self.linear.type(torch.cfloat) # TODO : is this trainable ?
		# return l(x)
		# l = self.linear.type(torch.cfloat) # TODO : is this trainable ?
		return self.linear(x)

class GELU(nn.Module):
	def __init__(self):
		super(GELU,self).__init__()
		self.σ = nn.GELU()
	def forward(self,x):
		if torch.is_complex(x): return self.σ(x.real) + 1j*       x.imag
		else:                   return self.σ(x)
		# x= torch.view_as_complex(x)
		# return self.σ(x.real) + 1j*       x.imag
		# return self.σ(x.real) + 1j*self.σ(x.imag)
		# return x/|x| * σ(|x|)

class MSE(nn.Module):
	def __init__(self):
		super(MSE,self).__init__()
	def forward(self,x,y):
		e=x-y
		return ((e*torch.conj(e)).real).mean()

# idea from https://discuss.pytorch.org/t/how-can-i-dropout-for-torch-complex-data-type/135577
# ! this should do the scaling of gradient from dropout
class Dropout(nn.Module):
	def __init__(self,p=0.5):
		super(Dropout,self).__init__()
		# self.do=nn.Dropout(p)
		self.p=p
	def forward(self,x):
		mask = torch.nn.functional.dropout(torch.ones_like(x.real),self.p,self.training)
		return x * mask

#############################################################
### USEFUL ℂ FUNCTIONS ######################################
#############################################################

class Smoothmax(nn.Module):
	def __init__(self,maximum,smoothness=1,mode="erf",inplace=False):
		super(Smoothmax,self).__init__()
		self.inplace = inplace
		s = 1/smoothness
		self.maximum = maximum
		if mode=="erf"     : self.λ = lambda x: .5+.5*torch.erf((x-self.maximum)*s)
		if mode=="sigmoid" : self.λ = lambda x: torch.sigmoid(2*(x-self.maximum)*s)
	def forward(self,x):
		if not self.inplace: x = x.clone()
		λ = self.λ(x.real)
		# new_real = (1-λ)*x.real + λ*self.maximum
		# return x - x.real + new_real
		return x - λ*x.real + λ*self.maximum

class Exp(nn.Module):
	def __init__(self,maximum=2**6,smoothness=1,inplace=False):
		super(Exp,self).__init__()
		self.inplace = inplace
		self.sm = Smoothmax(maximum,smoothness,inplace=True)
		if maximum==False: self.m = lambda x: x
		else:              self.m = lambda x: self.sm(x)
	def forward(self,x):
		if not self.inplace: x = x.clone()
		x = torch.exp(self.m(x))
		return x

class Log(nn.Module):
	def __init__(self,minimum=2**-8,smoothness=1,inplace=False):
		super(Log,self).__init__()
		self.inplace = inplace
		# if inplace: print("error : inplace not possible yet because of negative numbers and auto-ℂ conversion")
		self.sm = Smoothmax(minimum,-smoothness,inplace=True)
		if minimum==False: self.m = lambda x: x
		else:              self.m = lambda x: self.sm(x)
	def forward(self,x):
		if not self.inplace: x = x.clone()
		# x = x.clone()
		x = x.type(torch.cfloat)
		x = torch.log(self.m(x))
		return x

# a real Linear unit that evaluates on complex numbers
class CLinear(nn.Module):
	def __init__(self, dim_i, dim_o):
		super(CLinear, self).__init__()
		self.L = nn.Linear(dim_i, dim_o)
	def forward(self, x):
		real_output = self.L(x.real) # - self.L_imag(x.imag())
		imag_output = self.L(x.imag) # + self.L_imag(x.real())
		return real_output + 1j*imag_output

class ToComplex(nn.Module):
	def __init__(self): super(ToComplex,self).__init__()
	def forward(self,x): return x+0j

class ToReal(nn.Module):
	def __init__(self): super(ToReal,self).__init__()
	def forward(self,x): return x.real

# a real Linear unit that evaluates on complex numbers
class CConv2d(nn.Module):
	def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True, padding_mode='zeros', device=None, dtype=None):
		super(CConv2d, self).__init__()
		self.C = nn.Conv2d(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias, padding_mode, device, dtype)
	def forward(self, x):
		if torch.is_complex(x):
			real_output = self.C(x.real) # - self.L_imag(x.imag())
			imag_output = self.C(x.imag) # + self.L_imag(x.real())
			return real_output + 1j*imag_output
		else:
			return self.C(x)

# TODO
# class StochasticPool2d(nn.Module):
# 	def __init__(self, kernel_size,inplace=False):
# 		super(StochasticPool2d, self).__init__()
# 		self.kernel_size = kernel_size
# 		self.inplace = inplace
# 	def forward(self, x):
# 		if not self.inplace: x = x.clone()
# 		return x

#############################################################
### NN FUNCTIONS ############################################
#############################################################

class Pad(nn.Module):
	def __init__(self,*args):
		super().__init__()
		self.args=args
	def forward(self,x):
		return nn.functional.pad(x,*self.args)

class Stack(nn.Module):
	def __init__(self,activations=[]):
		super(Stack,self).__init__()
		self.σs = nn.ModuleList(activations)
		self.m = len(activations) # how many layers are stacked ?
	def forward(self,x):
		out = torch.hstack([ σ(x) for i,σ in enumerate(self.σs)])
		return out

def make_nn_fully_connected_complex(dim_i,dim_o,layer_count,σs=[GELU(),Exp()],dropout=0.5,after=[],before=[]):
	dims = np.linspace(dim_i,dim_o,layer_count+1)
	σ = Stack(σs)
	return nn.Sequential(*(
		before
		+
		[
			nn.Linear(int(λ)*(1 if index==0 else σ.m),int(μ),dtype=torch.cfloat) if mode == 0 else
			# nn.Linear(int(λ),int(μ)) if mode == 0 else
			σ if mode == 1 else 
			Dropout(dropout) if not index==(len(dims)-2) else nn.Identity() # Let's not put a dropout on the last layer
			for index,(λ,μ) in enumerate(zip(dims[:-1],dims[1:]))
			for mode in range(3)
		][:-1]
		+
		after
	))

def make_nn_fully_connected_real(dim_i,dim_o,layer_count,σ=nn.GELU(),after=[],before=[],dropout=0.5):
	dims = np.linspace(dim_i,dim_o,layer_count+1)
	return nn.Sequential(*(
		before
		+
		[
			nn.Linear(int(λ),int(μ)) if mode == 0 else
			σ                        if mode == 1 else
			nn.Dropout(dropout)
			for λ,μ in zip(dims[:-1],dims[1:])
			for mode in range(3)
		][:-1]  # Let's not put a dropout on the last layer
		+
		after
	))

# class Conv2d(nn.Module):
# 	def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1):
# 		super(Conv2d, self).__init__()
# 		def pad_compensator(kernel_size, stride, dilation):
# 			return ((stride - 1) + dilation * (kernel_size - 1)) // 2
# 		padding = padding + pad_compensator(kernel_size, stride, dilation)
# 		self.conv = nn.Conv2d(in_channels, out_channels, kernel_size, stride, padding=padding, dilation=dilation, groups=groups)
# 	def forward(self,x):
# 		return self.conv(x)

# for k in [1,2,3,4,5,6]:
# 	print(
# 		(Conv2d(1, 32, k, stride=1, padding=0, dilation=1, groups=1)(torch.randn(10,1,28,28))).shape
# 	)

def make_nn_cnn_real(channels,ker_size,pad=0,σs=[nn.Identity(),nn.GELU()],after=[],before=[],dropout=0.5):
	# pad = (ker_size-1)//2 if preserve else 0
	# pad_1 = ker_size // 2 + (ker_size - 2 * (ker_size // 2)) - 1
	# pad_2 = ker_size // 2
	# pad = (ker_size-1) // 2
	# print( pad )
	# pad = reduce(__add__,  [(k // 2 + (k - 2 * (k // 2)) - 1, k // 2) for k in ker_size[::-1]])
	# pad = 0
	σ = Stack(σs)
	return nn.Sequential(*
		before+[
			# nn.Sequential(nn.Conv2d(c1,c2,ker_size,padding=pad),σ,nn.Dropout(dropout))
			nn.Sequential(CConv2d(c1,c2,ker_size,padding=pad),σ,Dropout(dropout))
			for c1,c2 in [[channels[0],channels[1]]]
		]+[
			# nn.Sequential(nn.Conv2d(c1*σ.m,c2,ker_size,padding=pad),σ,nn.Dropout(dropout))
			nn.Sequential(CConv2d(c1*σ.m,c2,ker_size,padding=pad),σ,Dropout(dropout))
			for c1,c2 in zip(channels[1:-2],channels[2:-1])
		]+[
			# nn.Conv2d(channels[-2]*σ.m,channels[-1],ker_size,padding=pad), # nn.Dropout(dropout),  # Let's not put a dropout on the last layer
			CConv2d(channels[-2]*σ.m,channels[-1],ker_size,padding=pad), # nn.Dropout(dropout),  # Let's not put a dropout on the last layer
		]+after
	)

# for k,p in zip([1,2,3,4,5,6],[0,0,1,1,2,2]):
# 	print(make_nn_cnn_real([1,2,3],k,pad=p)(torch.randn(10,1,28,28)).shape)

# make a module out of a function
def modulize(fun):
	class Modulized(nn.Module):
		def __init__(self): super(Modulized,self).__init__()
		def forward(self, x): return fun(x)
	return Modulized()

# forward  pass is computed with fun
# backward pass is computed by differentiating app
def approximator(fun,app):
	class ApproximatorFunction(torch.autograd.Function):
		@staticmethod
		def forward(ctx, *args):
			ctx.save_for_backward(*args)
			return fun(*args)
		@staticmethod
		def backward(ctx, grad_output):
			args = ctx.saved_tensors
			torch.set_grad_enabled(True) # vjp's output require_grad
			y = torch.autograd.functional.vjp(app, args, v=grad_output)
			torch.set_grad_enabled(False)
			return y[1]
	return modulize(ApproximatorFunction.apply)

# forward  pass is computed with fun
# backward pass is computed using the provided jacobian
def jacobiator(fun,jacobian):
	class JacobiatorFunction(torch.autograd.Function):
		@staticmethod
		def forward(ctx, x):
			ctx.save_for_backward(x) # save input for backward pass
			return fun(x)
		@staticmethod
		def backward(ctx, grad_out):
			if not ctx.needs_input_grad[0]: return None                    # if grad not required, don't compute
			x, = ctx.saved_tensors                                         # restore input from context
			grad_input = torch.einsum('bj,bij->bi', grad_out, jacobian(x)) # grad_out.T ⋅ J(x)
			return grad_input
	return modulize(JacobiatorFunction.apply)

# # forward  pass is computed with fun
# # backward pass is computed directly with grad_fun
# def gradiator(fun,grad_fun):
# 	class GradiatorFunction(torch.autograd.Function):
# 		@staticmethod
# 		def forward(ctx, x):
# 			ctx.save_for_backward(x) # save input for backward pass
# 			return fun(x)
# 		@staticmethod
# 		def backward(ctx, grad_output):
# 			if not ctx.needs_input_grad[0]: return None # if grad not required, don't compute
# 			x, = ctx.saved_tensors # restore input from context
# 			grad_input = grad_fun(x,grad_output)
# 			return grad_input
# 	return modulize(GradiatorFunction.apply)

# remark 1 : as an analogy with openGL
#   x is intended to be used as "varying"
#   θ is intended to be used as "uniform"
# this is because we expcet internal parameters θ to be constant throughout many samples x,
# so one can use `einsum` to avoid pointless huge memory allocation (and obvious memory allocation limitation)
def gradiator(φ,grad_φ):
	class GradiatorFunction(torch.autograd.Function):
		@staticmethod
		def forward(ctx,x,θ):
			y = φ(x,θ)
			ctx.save_for_backward(x,θ,y) # save input for backward pass
			return y
		@staticmethod
		def backward(ctx, grad_o):
			# if not ctx.needs_input_grad[0]: return None # if grad not required, don't compute
			x,θ,y = ctx.saved_tensors # restore input from context
			grad_i = grad_φ(x,θ,y,grad_o)
			return grad_i
	return GradiatorFunction.apply

#############################################################

# class Gradiator_model(nn.Module):
# 	def __init__(self, dim_in: int, dim_out: int, custom_net=None):
# 		super(Gradiator_model, self).__init__()
# 		self.dim_in=dim_in
# 		self.dim_out=dim_out
# 		net = custom_net if not custom_net==None else nn.Sequential(
# 			nn.Linear(dim_in+dim_out       ,dim_in+int(dim_out/2)), nn.GELU(),
# 			nn.Linear(dim_in+int(dim_out/2),dim_in+int(dim_out/4)), nn.GELU(),
# 			nn.Linear(dim_in+int(dim_out/4),dim_in+int(dim_out/8)), nn.GELU(),
# 			nn.Linear(dim_in+int(dim_out/8),dim_in               ),
# 		).to(device)
# 		self.net = [net] # a hack to hide the model to self.parameters()
# 		self.opti = torch.optim.Adam(self.net[0].parameters(),lr=1e-2) # for learning it.
# 		self.dataset = [] # TODO : n-tree
# 	def sample (self, x, f_x     ): self.dataset.append([x,f_x])
# 	def forward(self, x, grad_out): return self.net[0](x, grad_out) # = grad_in
# 	def train(self, fun, epochs=256, sample_per_epoch=None, generator_mode=1, loss_mode=1): # train to properly model the function we are a jacobian of
# 		loss_hist=[]
# 		N = sample_per_epoch if not sample_per_epoch==None else dim_in * dim_out
# 		self.opti = torch.optim.Adam(self.net[0].parameters(),lr=1e-2) # for learning it.
# 		for n in range(epochs):
# 			if generator_mode==1:
# 				x1 = torch.randn(N,self.dim_in).to(device)
# 				x2 = torch.randn(N,self.dim_in).to(device)
# 				x = (x1+x2)/2 ; ε = x2-x1                  # compute midpoint and displacement
# 			elif generator_mode==2:
# 				x1 = torch.randn(N,self.dim_in).to(device)
# 				x2 = torch.randn(N,self.dim_in).to(device)
# 				x = x1 ; ε = x2-x1                      # compute midpoint and displacement
# 			else:
# 				x = torch.randn(N,self.dim_in).to(device)      # generate midpoint
# 				ε = torch.randn(N,self.dim_in).to(device)/100. # generate displacement
# 				x1=x-ε/2; x2=x+ε/2                             # compute a pair
# 			y1=fun(x1); y2=fun(x2) # shape=(N,dim_out), note that given y's computation is supposed to be super cheap, we don't need to recycle samples and store
# 			diff = (y2-y1)
# 			self.opti.zero_grad()
# 			loss=0
# 			if loss_mode==1:
# 				for i in range(self.dim_out):
# 					g_out = F.one_hot(i*torch.ones(N).to(torch.int64),self.dim_out).float().to(device)
# 					loss += nn.MSELoss()(torch.einsum('bi,bi->b', self(x,g_out), ε), torch.einsum('bi,bi->b', g_out, diff))
# 			elif loss_mode==2:
# 				g_out = F.one_hot(torch.arange(N)%self.dim_out,self.dim_out).float().to(device)
# 				loss += nn.MSELoss()(torch.einsum('bi,bi->b', self(x,g_out), ε), torch.einsum('bi,bi->b', g_out, diff))
# 			loss.backward()
# 			loss_hist.append(loss.item())
# 			self.opti.step()
# 		plt.plot(np.log(loss_hist)/np.log(10))

class NN_ADD_PARAM(nn.Module):
	def __init__(self,model,θ_shape=0):
		super(NN_ADD_PARAM,self).__init__()
		self.model = model
		self.θ = nn.parameter.Parameter(torch.randn(θ_shape))
	def forward(self,x):
		θθ = self.θ
		if len(x.shape) == 2: # if multiple samples, reshape θ accordingly
			sample_count = x.shape[0]
			θθ = self.θ.repeat(sample_count,1)
		X = torch.hstack((θθ,x)) # actual usage 
		return self.model(X)

def train_approx(approx,fun,epochs,batch_generator,lr=1e-4):
	if not hasattr(approx, 'my_optimizer'):
		approx.my_optimizer = torch.optim.Adam(approx.parameters(), lr=1e-4)
	approx_opti = approx.my_optimizer
	optim_lr(approx_opti,lr)
	plt.figure(figsize=(20,5))
	plt.xscale('log')
	for epoch in range(epochs): # it's more like sample count as dataset is generated live
		x = batch_generator().to(device)
		approx.train()
		approx_opti.zero_grad()
		loss = F.mse_loss(approx(x), fun(x))
		loss.backward()
		approx_opti.step()
		plt.scatter(epoch,np.log(loss.item()),s=2)
# batch_size = 1024
# fun_input_dim=20
# fun = lambda x: nn.Linear(20,20)(x)
# approx = some_neural_net
# train_approx(approx,fun,epochs,batch_generator=lambda:3*torch.randn([batch_size]+fun_input_dim),lr=1e-4)

def nn_dropper(n: int): return modulize(lambda x : x[:,:n])

def nn_permute_and_drop(input_size: int, output_size: int):
	perm = torch.randperm(input_size)
	return modulize(lambda x : x[:,perm][:,:output_size])

#############################################################
### USEFUL FUNCTIONS ########################################
#############################################################

# def smoothmax(x,max): return (max/2)*4*(torch.sigmoid(x*2/max)-.5)

# def demo_smoothmax():
#   x = torch.linspace(-0,100,100)
#   plt.plot(x,x)
#   plt.plot(x,smoothmax(x,20))
#   plt.plot(x,20*torch.ones(x.shape))


### interpolations ########################
def nearest_floor(t,v1,v2): return v1
def nearest(t,v1,v2): λ = (t>0.5).float(); return v1*(1-λ) + v2*λ
def lerp(t,v1,v2):                         return v1*(1-t) + v2*t
def smoothstep(t,v1,v2): λ = t*t*(3-2*t);  return v1*(1-λ) + v2*λ

### sampler ###############################
def sampler1d_repeat(tex,x,interpolation=lerp):
	l = len(tex)
	X=x*l # so that x ∈ [0,1] wil cover the whole texture
	Xi = X.int() - (X<=0)*1
	t = X-Xi
	return interpolation(t, tex[Xi % l], tex[(Xi + 1) % l])

def sampler1d_clamp(tex,x,interpolation=lerp):
	l = len(tex)
	X=x*l # so that x ∈ [0,1] wil cover the whole texture
	Xi = X.int() - (X<=0)*1
	t = X-Xi
	return interpolation(t, tex[torch.clamp(Xi,0,l-1)], tex[torch.clamp(Xi+1,0,l-1)])

def sampler2d_nearest_clamp(tex, coord):
	c, h, w = tex.shape
	# remap coord from [0,1]x[0,1] to [0,h]x[0,w]
	coord[:, 0] *= h-1
	coord[:, 1] *= w-1
	coord = torch.round(coord).long()
	coord_x = torch.clamp(coord[:, 1],0,w-1) # repeat last pixel
	coord_y = torch.clamp(coord[:, 0],0,h-1) # repeat last pixel
	pixels = tex[:, coord_y, coord_x]
	return pixels.permute(1,0)

### polynoms ##############################
def poly(coeff): # each components get a polynome applied
	N = range(len(coeff))
	return lambda x: sum([coeff[n]*(x**n) for n in N])

### fake gradient | cantor ################
def cantor_function(x,n=6):
	if n==0: return x
	return (
		(((0 <=x) & (x<=1/3)) * cantor_function(3*x,n-1)/2) +
		((1/3< x) & (x< 2/3)) * (1/2) +
		(2/3 <=x)             * (1/2+cantor_function(3*x-2,n-1)/2)
	)

def cantor():
	if not hasattr(cantor, 'array'):
		cantor.array = cantor_function(torch.linspace(0,1,3**5+1),5).to(device)
	return modulize(lambda x: sampler1d_clamp(cantor.array,x,nearest_floor))

def cantor2():
	if not hasattr(cantor, 'array'):
		cantor.array = cantor_function(torch.linspace(0,1,3**5+1),5).to(device)
	return modulize(gradiator(
		lambda x,θ=None: torch.floor(x) + sampler1d_repeat(cantor.array,x,nearest_floor),
		lambda x,θ,y,grad: grad # fake the gradient to identity
	))

### fake gradient | sigmoid staircase #####
### fake gradient | sigmoid sinus #########
### fake gradient | binary sinus ##########

### heterogenous ##########################
def multi(σs):
	def multi_(x):
		σcount = len(σs)
		out = torch.zeros_like(x)
		for (i,σ) in enumerate(σs):
			index = torch.arange(i,len(x)-1,σcount)
			out[index] = σ(x[index])
		return out
	return modulize(multi_)

def random_function_light(dim_in=196,dim_out=196,device=device):
	L1 = nn_freeze(nn.Linear(dim_in,dim_in)).to(device)
	L2 = nn_freeze(nn.Linear(dim_in,dim_out)).to(device)
	L3 = nn_freeze(nn.Linear(dim_in,dim_out)).to(device)
	L2.weight.data = torch.relu(L2.weight.data)
	def f(x):
		x = torch.sigmoid(L2(torch.sin(L1(x)) + torch.sin(torch.relu(L2(x))))) + torch.exp(-(L3(torch.sin(x))**2))
		# x = torch.sigmoid(torch.sin(L1(x)) + torch.relu(L2(x))) + torch.exp(-(L3(torch.sin(x))**2))
		# x = L1(x) + torch.relu(L2(x))
		return torch.sin(x)
		# return torch.sigmoid(x)
	# φ = (1 + np.sqrt(5))/2
	# # def fractale(x): return f(x) + f(x*2)/2 + f(x*4)/4 + f(x*8)/8
	# def fractale(x): return f(x/φ) + f(x+φ*π) + f((x+π)*φ)/2 + f(-(x-2*φ)*φ*π)/4 + f(x*φ*φ*φ)/8
	return modulize(f).to(device)

def random_function_lighter(dim=196,device=device):
	G = torch.zeros((dim,dim)).to(device)
	for i in range(2):
		G += torch.diag(torch.randn(dim-i)/(i+1),+i).to(device)
		G += torch.diag(torch.randn(dim-i)/(i+1),-i).to(device)
	L1 = nn_freeze(nn.Linear(dim,dim)).to(device)
	# L1.weight.data = L1.weight.data * G
	def f(x):
		x = L1(x)
		# return x
		return torch.relu(x)
		# return torch.sin(x)
		# return torch.sigmoid(x)
	φ = (1 + np.sqrt(5))/2
	# def fractale(x): return f(x) + f(x*2)/2 + f(x*4)/4 + f(x*8)/8
	def fractale(x): return f(x/φ) + f(x+φ*π) + f((x+π)*φ)/2 + f(-(x-2*φ)*φ*π)/4 + f(x*φ*φ*φ)/8
	# return modulize(fractale).to(device)
	return modulize(f).to(device)
